
/* $Header: /cvs/faces/faces/faces/face_socket.c,v 1.6 2002/03/03 17:37:45 devface Exp $
 *
 * This module is a massivily modified version of mutt_socket.c
 * found in the mutt mailreader.  The origional copyrights read:
 *
 * Copyright (C) 1998 Michael R. Elkins <me@cs.hmc.edu>
 * Copyright (C) 1999-2001 Brendan Cully <brendan@kublai.com>
 * Copyright (C) 1999-2000 Tommi Komulainen <Tommi.Komulainen@iki.fi>
 * 
 * Mutt is distributed under the FSF GNU General Public Licence.  The
 * complete Mutt program itself can be found at http://www.mutt.org/.
 *
 * This version is
 *  Copyright (C) 2001-2002  Robert Adams (misterblue@misterblue.com)
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 * 
 *  No responsibility is taken for any errors on inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 */ 

#include "faces.h"
#include "face_socket.h"

#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>

#ifdef SDEBUG
#define  S_DEBUG(stmt)  stmt
#else
#define  S_DEBUG(stmt)
#endif /*SDEBUG*/

/* forward declarations */
static int face_socket_preconnect(void);
static int face_socket_connect(int fd, struct sockaddr* sa);


/* Allocate and initialize the datastructure that keeps
 * the state of the socketconnection.
 */

CONNECTION *
build_connection(char* host, int port)
{
    CONNECTION* conn;

    conn = (CONNECTION *) Malloc(sizeof(CONNECTION));
    CLEARMEM(conn, sizeof(CONNECTION));
    conn->bufpos = 0;
    conn->fd = -1;
    conn->available = 0;
    conn->read = raw_socket_read;
    conn->write = raw_socket_write;
    conn->open = raw_socket_open;
    conn->close = raw_socket_close;
    conn->host = strdup(host);
    conn->port = port;

    return(conn);
}


/* Release a connection */

int 
free_connection(CONNECTION* conn)
{
    if (conn) {
        if (conn->fd >= 0) {
            (void)face_socket_close(conn);
        }
        if (conn->host) {
            FREE(conn->host);
            conn->host = NULL;
        }
        FREE(conn);
    }

    return(0);
}


/* Simple read buffering to speed things up. */

int
face_socket_readchar(CONNECTION *conn, char *c)
{
    if (conn->bufpos >= conn->available) {
        if (conn->fd >= 0) {
            conn->available = conn->read(conn, conn->inbuf, 
                                         sizeof(conn->inbuf));
        } else {
            FPRINTF(stderr, 
             "face_socket_readchar: attempt to read from closed connection.\n");
            return(-1);
        }
        conn->bufpos = 0;
        if (conn->available == 0) {
            S_DEBUG(FPRINTF(stderr, "Connection to %s closed\n", conn->host));
        }
        if (conn->available <= 0) {
            face_socket_close (conn);
            return(-1);
        }
    }
    *c = conn->inbuf[conn->bufpos];
    conn->bufpos++;

    return(1);
}


/* Read a line from the connection.  Lines start at the current
 * position and end with a line-feed.  If a CR occurs just before
 * the LF, it is removed (ie, CRLF --> LF).
 * Returns the number of bytes read (NOT the length of the returned string)
 * Returns -1 if the connection ended without terminating the line.
 */

int 
face_socket_readln(CONNECTION* conn, char* buf, size_t buflen)
{
    char ch;
    int i;

    /* fetch characters and build output line */
    for (i = 0; i < buflen-1; i++) {
        if (face_socket_readchar(conn, &ch) != 1) {
            buf[i] = '\0';
            return(-1);
        }

        if (ch == '\n') {
            break;
        }
        buf[i] = ch;
    }

    /* strip \r from \r\n termination */
    if (i && buf[i-1] == '\r') {
        buf[--i] = '\0';
    } else {
        buf[i] = '\0';
    }

    S_DEBUG(FPRINTF(stderr, "< '%s'\n", buf));

    /* number of bytes read, not strlen */
    return(i + 1);
}


/* Do any before-using-sockets action 
 * (setting up ssh or tunnel, for instance)
 * Return 0 on success, error code otherwise.
 */

static int 
face_socket_preconnect(void)
{
    /* we don't have anything here */

    return(0);
}


/* Wrappers */

int 
face_socket_open(CONNECTION* conn) 
{
    if (face_socket_preconnect()) {
        return(-1);
    }

    return(conn->open(conn));
}


int 
face_socket_close(CONNECTION* conn)
{
    int rc = -1;

    if (conn->fd < 0) {
        S_DEBUG(FPRINTF(stderr, 
                   "face_socket_close: Attempt to close closed connection.\n"));
    } else {
        rc = conn->close(conn);
    }

    conn->fd = -1;

    return(rc);
}


int 
face_socket_read(CONNECTION* conn, char* buf, size_t len)
{
    int rc;

    if (conn->fd < 0) {
        FPRINTF(stderr, 
                "face_socket_read: attempt to read from closed connection\n");
        return(-1);
    }

    rc = conn->read(conn, buf, len);
    /* EOF */
    if (rc == 0) {
        FPRINTF(stderr, "Connection to %s closed", conn->host);
    }
    if (rc <= 0) {
        face_socket_close(conn);
    }

    return(rc);
}


int 
face_socket_write(CONNECTION *conn, const char *buf, int len)
{
    int rc;

    S_DEBUG(FPRINTF(stderr, "> '%s'", buf));

    if (conn->fd < 0) {
        FPRINTF(stderr, 
                "face_socket_write: attempt to write to closed connection\n");
        return(-1);
    }

    if ((rc = conn->write(conn, buf, len)) < 0) {
        FPRINTF(stderr, "face_socket_write: error writing, closing socket\n");
        face_socket_close (conn);

        return(-1);
    }

    if (rc < len) {
        FPRINTF(stderr, 
                "face_socket_write: ERROR: wrote %d of %d bytes!\n", rc, len);
    }

    return(rc);
}


/* the real operations on the sockets */

int 
raw_socket_close(CONNECTION *conn)
{
    return(close(conn->fd));
}


int 
raw_socket_read(CONNECTION* conn, char *buf, size_t len)
{
    int rc;

    if ((rc = read(conn->fd, buf, len)) == -1) {
        FPRINTF(stderr, "Error talking to %s (%s)", conn->host,
                strerror(errno));
    }

    return(rc);
}


int 
raw_socket_write(CONNECTION* conn, const char *buf, size_t count)
{
    int rc;

    if ((rc = write(conn->fd, buf, count)) == -1) {
        FPRINTF(stderr, "Error talking to %s (%s)", conn->host,
                strerror(errno));
    }

    return(rc);
}


int 
raw_socket_open(CONNECTION* conn)
{
    int rc;
    int fd;
    char buf[LONG_STRING];

#ifdef HAVE_GETADDRINFO
/* --- IPv4/6 --- */

    /* "65536\0" */
    char port[6];
    struct addrinfo hints;
    struct addrinfo* res;
    struct addrinfo* cur;

    /* we accept v4 or v6 STREAM sockets */
    memset(&hints, 0, sizeof (hints));

    if (option (OPTUSEIPV6)) {
        hints.ai_family = AF_UNSPEC;
    } else {
        hints.ai_family = AF_INET;
    }

    hints.ai_socktype = SOCK_STREAM;

    snprintf(port, sizeof (port), "%d", conn->port);

    face_message(_("Looking up %s...\n"), conn->host);

    rc = getaddrinfo(conn->host, conn->port, &hints, &res);
    if (rc) {
        FPRINTF(stderr, "Could not find the host \"%s\"\n", conn->host);
        return(-1);
    }

    face_message(_("Connecting to %s...\n"), conn->host); 

    rc = -1;
    for (cur = res; cur != NULL; cur = cur->ai_next) {
        fd = socket(cur->ai_family, cur->ai_socktype, cur->ai_protocol);
        if (fd >= 0) {
            if ((rc = socket_connect(fd, cur->ai_addr)) == 0) {
                fcntl(fd, F_SETFD, FD_CLOEXEC);
                conn->fd = fd;
                break;
            } else {
                close(fd);
            }
        }
    }

    freeaddrinfo(res);

#else
  /* --- IPv4 only --- */

    struct sockaddr_in sin;
    struct hostent* he;
    int i;

    memset(&sin, 0, sizeof (sin));
    sin.sin_port = htons (conn->port);
    sin.sin_family = AF_INET;

    S_DEBUG(FPRINTF(stderr, "Looking up %s...\n", conn->host));

    if ((he = gethostbyname(conn->host)) == NULL) {
        FPRINTF(stderr, "Could not find the host \"%s\"\n", conn->host);

        return(-1);
    }

    S_DEBUG(FPRINTF(stderr, "Connecting to %s...\n", conn->host));

    rc = -1;
    for (i = 0; he->h_addr_list[i] != NULL; i++) {
        memcpy(&sin.sin_addr, he->h_addr_list[i], he->h_length);
        fd = socket(PF_INET, SOCK_STREAM, IPPROTO_IP);

        if (fd >= 0) {
            if ((rc = face_socket_connect(fd, (struct sockaddr*) &sin)) == 0) {
                S_DEBUG(FPRINTF(stderr,"Connection successful\n"));
                conn->fd = fd;
                break;
            } else {
                close(fd);
            }
        }
    }

#endif
    if (rc) {
        SPRINTF(buf, "Could not connect to %s (%s).", conn->host,
                (rc > 0) ? strerror(rc) : "unknown error");
        display_error(buf);
        return(-1);
    }

    return(0);
}


/* face_socket_connect: set up to connect to a socket fd. */

static int 
face_socket_connect(int fd, struct sockaddr* sa)
{
    int sa_size;
    int save_errno;

    if (sa->sa_family == AF_INET)
        sa_size = sizeof(struct sockaddr_in);
#ifdef HAVE_GETADDRINFO
    else if (sa->sa_family == AF_INET6)
        sa_size = sizeof(struct sockaddr_in6);
#endif
    else {
        FPRINTF(stderr, "Unknown address family!\n");
        return(-1);
    }
  
    save_errno = 0;

    if (connect(fd, sa, sa_size) < 0) {
        save_errno = errno;
        FPRINTF(stderr, "Connection failed. errno: %d...\n", errno);
    }

    return(save_errno);
}
