
/* $Header: /cvs/faces/faces/faces/parsedate.c,v 1.4 2002/03/03 17:37:45 devface Exp $
 *
 * This module is a massively modified version of mail.c from Mark
 * Crispin's imap2000 toolkit.  The original copyrights read:
 *
 * Copyright (C) 2001 University of Washington
 * 
 * This version is
 *  Copyright (C) 2001-2002  Robert Adams (misterblue@misterblue.com)
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 * 
 *  No responsibility is taken for any errors on inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 */

#include "faces.h"
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "extern.h"

/* Mail parse date into a long
 * Accepts: date line to parse
 * Returns: time in a long  parse successful, else zero
 * This routine parses dates as follows:
 * . leading three alphas followed by comma and space are ignored
 * . date accepted in format: mm/dd/yy, mm/dd/yyyy, dd-mmm-yy, dd-mmm-yyyy,
 *    dd mmm yy, dd mmm yyyy
 * . two and three digit years interpreted according to RFC 2822 rules
 * . space or end of string required
 * . time accepted in format hh:mm:ss or hh:mm
 * . end of string accepted
 * . timezone accepted: hyphen followed by symbolic timezone, or space
 *    followed by signed numeric timezone or symbolic timezone
 * Examples of normal input:
 * . IMAP date-only (SEARCH):
 *    dd-mmm-yyyy
 * . IMAP date-time (INTERNALDATE):
 *    dd-mmm-yyyy hh:mm:ss +zzzz
 * . RFC-822:
 *    www, dd mmm yy hh:mm:ss zzz
 * . RFC-2822:
 *    www, dd mmm yyyy hh:mm:ss +zzzz
 */

#define MAILTMPLEN 60

#ifdef DATEDEBUG
#define  D_DEBUG(stmt)  stmt
#else
#define  D_DEBUG(stmt)
#endif /*DATEDEBUG*/


time_t 
parsedate(char *ss)
{
    int ms, mi;
    struct tm t;
    time_t tn;
    int zoccidentf = 0;
    int zoccidenth = 0;
    int zoccidentm = 0;
    char *s;
    char tmp[MAILTMPLEN];

    CLEARMEM(&t, sizeof(t));    /* no time in the beginning */

    /* make a writeable uppercase copy */
    if (ss && *ss && (strlen (ss) < (size_t)MAILTMPLEN))  {
        s = strcpy(tmp, ss);
    } else  {
        return(0);
    }

    /* make it all upper case */
    while (*s) {
        *s = toupper(*s);
        s++;
    }
    s = tmp;
    D_DEBUG(FPRINTF(stderr, "Date line = '%s'\n", s));

    if (strncmp(s, "DATE:", 5) != 0) {
/* This doesn't make a good Date line if it doesn't start with "Date:" */
        D_DEBUG(FPRINTF(stderr, "Not legal RFC822 date line\n"));
        return(0);
    }
    s += 5;
    while (*s == ' ')  {
        s++;
    }

    tn = 0;
    /* skip over possible day of week */
    if (isalpha((int) *s) && 
        isalpha((int) s[1]) && 
        isalpha((int) s[2]) && 
        (s[3] == ',') && (s[4] == ' ')) {
        s += 5;
    }
    /* parse first number (probable month) */
    while (*s == ' ') {
        s++;
    }
    if (!(t.tm_mon = strtoul((const char *) s, &s, 10)))  {
        return(0);
    }

    switch (*s) {            /* different parse based on delimiter */
        case '/':            /* mm/dd/yy format */
            if (isdigit((int) *++s) 
                    && (t.tm_mday = strtoul((const char *) s,&s,10)) 
                    && (*s == '/') && isdigit((int) *++s)) {
                t.tm_year = strtoul((const char *) s, &s, 10);
                if (*s == '\0') {
                    break;    /* must end here */
                }
            }
            return(0);            /* bogon */

        case ' ':            /* dd mmm yy format */
            while (s[1] == ' ') {
                /* slurp extra whitespace */
                s++;
            }

        case '-':            /* dd-mmm-yy format */
            t.tm_mday = t.tm_mon;            /* so the number we got is a day */
            /* make sure string long enough! */
            if (strlen(s) < (size_t) 5)  {
                return(0);
            }
/* Some compilers don't allow `<<' and/or longs in case statements. */

            /* slurp up the month string */
            ms = ((s[1] - 'A') * 1024) + ((s[2] - 'A') * 32) + (s[3] - 'A');
            switch (ms) {        /* determine the month */
                case (('J'-'A') * 1024) + (('A'-'A') * 32) + ('N'-'A'):
                    t.tm_mon = 1;
                    break;

                case (('F'-'A') * 1024) + (('E'-'A') * 32) + ('B'-'A'):
                    t.tm_mon = 2;
                    break;

                case (('M'-'A') * 1024) + (('A'-'A') * 32) + ('R'-'A'):
                    t.tm_mon = 3; 
                    break;

                case (('A'-'A') * 1024) + (('P'-'A') * 32) + ('R'-'A'): 
                    t.tm_mon = 4;
                    break;

                case (('M'-'A') * 1024) + (('A'-'A') * 32) + ('Y'-'A'): 
                    t.tm_mon = 5; 
                    break;

                case (('J'-'A') * 1024) + (('U'-'A') * 32) + ('N'-'A'): 
                    t.tm_mon = 6; 
                    break;

                case (('J'-'A') * 1024) + (('U'-'A') * 32) + ('L'-'A'): 
                    t.tm_mon = 7; 
                    break;

                case (('A'-'A') * 1024) + (('U'-'A') * 32) + ('G'-'A'): 
                    t.tm_mon = 8; 
                    break;

                case (('S'-'A') * 1024) + (('E'-'A') * 32) + ('P'-'A'): 
                    t.tm_mon = 9; 
                    break;

                case (('O'-'A') * 1024) + (('C'-'A') * 32) + ('T'-'A'): 
                    t.tm_mon = 10; 
                    break;

                case (('N'-'A') * 1024) + (('O'-'A') * 32) + ('V'-'A'): 
                    t.tm_mon = 11; 
                    break;

                case (('D'-'A') * 1024) + (('E'-'A') * 32) + ('C'-'A'): 
                    t.tm_mon = 12; 
                    break;

                default:
                    return(0);    /* unknown month */
            }

            if (s[4] == *s) {
                s += 5;           /* advance to year */
            } else {              /* first three were OK, possibly full name */
                mi = *s;          /* note delimiter, skip alphas */
                for (s += 4; isalpha((int) *s); s++);
                /* error if delimiter not here */
                if (mi != *s++)  {
                    return(0);
                }
            }

            while (*s == ' ') {
                s++;    /* parse year */
            }
            if (isdigit((int) *s)) {        /* must be a digit here */
                t.tm_year = strtoul((const char *) s, &s, 10);
                if (*s == '\0' || *s == ' ') {
                    break;
                }
            }

        default:
            return(0);            /* unknown date format */
    }

    /* minimal validity check of date */
    if ((t.tm_mday > 31) || (t.tm_mon > 12))  {
        return(0); 
    }
    if (t.tm_year < 49) {
        /* RFC 2282 rules for two digit years 00-49 */
        t.tm_year += 2000;
    } else {
        /* 2-digit years 50-99 and 3-digit years */
        if (t.tm_year < 999) {
            t.tm_year += 1900;
        }
    }
    if (t.tm_year < 1950) {
        return(0);    /* reject prehistoric year */
    }

    ms = '\0';            /* initially no time zone string */
    if (*s) {            /* time specification present? */
        /* parse time */
        t.tm_hour = strtoul((const char *) s+1, &s, 10);
        if (*s != ':')  {
            return(0);
        }
        t.tm_min = strtoul((const char *) ++s, &s, 10);
        t.tm_sec = (*s == ':') ? strtoul((const char *) ++s, &s, 10) : 0;
        /* validity check time */
        if ((t.tm_hour > 23) || (t.tm_min > 59) || (t.tm_sec > 60))  {
            return(0); 
        }

        /* set values in elt */
        switch (*s) {        /* time zone specifier? */
            case ' ':            /* numeric time zone */
                while (s[1] == ' ') {
                    s++;    /* slurp extra whitespace */
                }
                if (!isalpha ((int) s[1])) {    /* treat as '-' case if alphabetic */
                    /* test for sign character */
                    if ((zoccidentf = (*++s == '-')) || (*s == '+'))  {
                        s++;
                    }
                    /* validate proper timezone */
                    if (isdigit((int) *s) && 
                        isdigit((int) s[1]) && 
                        isdigit((int) s[2]) && 
                        (s[2] < '6') && 
                        isdigit((int) s[3])) {
                        zoccidenth = ((*s - '0') * 10 + (s[1] - '0'));
                        zoccidentm = (s[2] - '0') * 10 + (s[3] - '0');
                    }
                    break;        /* all done! */
                }
            /* falls through */

            case '-':            /* symbolic time zone */
                if (!(ms = *++s))  {
                    ms = 'Z';
                } else if (*++s) {        /* multi-character? */
                    /* yes, make compressed three-byte form */
                    ms -= 'A'; ms *= 1024;
                    ms += ((*s++ - 'A') * 32);
                    if (*s)  {
                        ms += *s++ - 'A';
                    }
                    if (*s)  {
                        ms = '\0';    /* more than three characters */
                    }
                }

            default:            /* ignore anything else */
                break;
        }
    }

    /*  This is not intended to be a comprehensive list of all possible
     * timezone strings.  Such a list would be impractical.  Rather, this
     * listing is intended to incorporate all military, North American, and
     * a few special cases such as Japan and the major European zone names,
     * such as what might be expected to be found in a Tenex format mailbox
     * and spewed from an IMAP server.  The trend is to migrate to numeric
     * timezones which lack the flavor but also the ambiguity of the names.
     *
     *  RFC-822 only recognizes UT, GMT, 1-letter military timezones, and the
     * 4 CONUS timezones and their summer time variants.  [Sorry, Canadian
     * Atlantic Provinces, Alaska, and Hawaii.]
    */

    switch (ms) {            /* determine the timezone */
        /* Universal */
        case (('U'-'A')*1024)+(('T'-'A')*32):
#ifndef STRICT_RFC822_TIMEZONES
        case (('U'-'A')*1024)+(('T'-'A')*32)+'C'-'A':
#endif
        /* Greenwich */
        case (('G'-'A')*1024)+(('M'-'A')*32)+'T'-'A':
        case 'Z': zoccidenth = 0; break;

        /* oriental (from Greenwich) timezones */
#ifndef STRICT_RFC822_TIMEZONES
        /* Middle Europe */
        case (('M'-'A')*1024)+(('E'-'A')*32)+'T'-'A':
#endif
#ifdef BRITISH_SUMMER_TIME
        /* British Summer */
        case (('B'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
        case 'A': zoccidenth = 1; break;
#ifndef STRICT_RFC822_TIMEZONES
        /* Eastern Europe */
        case (('E'-'A')*1024)+(('E'-'A')*32)+'T'-'A':
#endif
        case 'B': 
            zoccidenth = 2; 
            break;
        case 'C': 
            zoccidenth = 3; 
            break;
        case 'D': 
            zoccidenth = 4; 
            break;
        case 'E': 
            zoccidenth = 5; 
            break;
        case 'F': 
            zoccidenth = 6; 
            break;
        case 'G': 
            zoccidenth = 7; 
            break;
        case 'H': 
            zoccidenth = 8; 
            break;
#ifndef STRICT_RFC822_TIMEZONES
        /* Japan */
        case (('J'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
        case 'I': 
            zoccidenth = 9; 
            break;
        case 'K': 
            zoccidenth = 10; 
            break;
        case 'L': 
            zoccidenth = 11; 
            break;
        case 'M': 
            zoccidenth = 12; 
            break;
        /* occidental (from Greenwich) timezones */
        case 'N': 
            zoccidentf = 1; 
            zoccidenth = 1; 
            break;
        case 'O': 
            zoccidentf = 1; 
            zoccidenth = 2; 
            break;
#ifndef STRICT_RFC822_TIMEZONES
        case (('A'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
#endif
        case 'P': 
            zoccidentf = 1; 
            zoccidenth = 3; 
            break;
#ifdef NEWFOUNDLAND_STANDARD_TIME
        /* Newfoundland */
        case (('N'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
            zoccidentf = 1; 
            zoccidenth = 3; 
            zoccidentm = 30; 
            break;
#endif
#ifndef STRICT_RFC822_TIMEZONES
        /* Atlantic */
        case (('A'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
        /* CONUS */
        case (('E'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
        case 'Q': 
            zoccidentf = 1; 
            zoccidenth = 4; 
            break;
        /* Eastern */
        case (('E'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
        case (('C'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
        case 'R': 
            zoccidentf = 1; 
            zoccidenth = 5; 
            break;
        /* Central */
        case (('C'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
        case (('M'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
        case 'S': 
            zoccidentf = 1; 
            zoccidenth = 6; 
            break;
        /* Mountain */
        case (('M'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
        case (('P'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
        case 'T': 
            zoccidentf = 1; 
            zoccidenth = 7; 
            break;
        /* Pacific */
        case (('P'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#ifndef STRICT_RFC822_TIMEZONES
        case (('Y'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
#endif
        case 'U': 
            zoccidentf = 1; 
            zoccidenth = 8; 
            break;
#ifndef STRICT_RFC822_TIMEZONES
        /* Yukon */
        case (('Y'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
        case 'V': 
            zoccidentf = 1; 
            zoccidenth = 9; 
            break;
#ifndef STRICT_RFC822_TIMEZONES
        /* Hawaii */
        case (('H'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
        case 'W': 
            zoccidentf = 1; 
            zoccidenth = 10; 
            break;
        /* Nome/Bering/Samoa */
#ifdef NOME_STANDARD_TIME
        case (('N'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
#ifdef BERING_STANDARD_TIME
        case (('B'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
#ifdef SAMOA_STANDARD_TIME
        case (('S'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
#endif
        case 'X': 
            zoccidentf = 1; 
            zoccidenth = 11; 
            break;
        case 'Y': 
            zoccidentf = 1; 
            zoccidenth = 12; 
            break;
        default:            /* unknown time zones treated as local */
            break;
    }
    
    D_DEBUG(FPRINTF(stderr, "parsedate: %d-%02d-%02d %d:%02d:%02d %c%d%02d\n",
                t.tm_year, t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec,
                (zoccidentf?'-':'+'), zoccidenth, zoccidentm));
    /* make a time from the above parts */
	if (t.tm_year > 200) {
		t.tm_year -= 1900;
	}
    tn = mktime(&t);

    /* Modify the time for the time zone */
    if (zoccidentf) {
        zoccidenth *= -1;
        zoccidentm *= -1;
    }
    tn += (zoccidenth*60+zoccidentm)*60;

    return(tn);
}
